from os import cpu_count

import click as click

from openvariant.task.cat import cat as cat_task
from openvariant.task.count import count as count_task
from openvariant.task.groupby import group_by as group_by_task


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--debug', help='Show debug messages.', is_flag=True)
@click.version_option()
def cli(debug):
    pass


@click.command(name="cat", short_help='Concatenate files to standard input')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--where', '-w', type=click.STRING, default=None, help="Filter expression. eg: CHROMOSOME == 4")
@click.option('--annotations', '-a', default=None)
@click.option('--header', help="Show the result header", is_flag=True)
def cat(input_path: str, where: str or None, annotations: str or None, header: bool):
    cat_task(input_path, annotations, where, header)


@click.command(name="count", short_help='Number of rows that matches a specified criterion')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--where', '-w', multiple=False, type=click.STRING)
@click.option('--group_by', '-g', type=click.STRING)
@click.option('--annotations', '-a', default=None, type=click.Path(exists=True))
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
def count(input_path: str, where: str, group_by: str, cores: int, quite: bool, annotations: str or None) -> None:
    result = count_task(input_path, annotations, group_by=group_by, where=where, cores=cores, quite=quite)

    if len(result[1]) > 0:
        for k, v in sorted(result[1].items(), key=lambda res: res[1]):
            print("{}\t{}".format(k, v))

    print("TOTAL\t{}".format(result[0]))


@click.command(name="groupby", short_help='Groups rows that have the same values into summary rows')
@click.argument('input_path', type=click.Path(exists=True), default='.')
@click.option('--header', help='Send header as first row', is_flag=True)
@click.option('--show', help='Show group by each row', is_flag=True)
@click.option('--group_by', '-g', type=click.STRING)
@click.option('--where', '-w', type=click.STRING)
@click.option('--script', '-s', type=click.STRING)
@click.option('--annotations', '-a', default=None, type=click.Path(exists=True))
@click.option('--cores', help='Maximum processes to run in parallel.', type=click.INT, default=cpu_count())
@click.option('--quite', '-q', help="Don't show the progress, only the total count.", is_flag=True)
def groupby(input_path: str, script: str, where: str, group_by: str, cores: int, quite: bool, annotations: str,
            header: bool, show: bool):
    for group_key, group_result, command in group_by_task(input_path, annotations, script, key_by=group_by, where=where,
                                                          cores=cores, quite=quite, header=header):
        for r in group_result:
            if command:
                print(f"{group_key}\t{r}") if show else print(f"{r}")
            else:
                if header:
                    print(f"{r}")
                    header = False
                else:
                    print(f"{group_key}\t{r}") if show else print(f"{r}")


cli.add_command(cat)
cli.add_command(count)
cli.add_command(groupby)

if __name__ == "__main__":
    cli()



import csv
import ctypes
import gzip
import lzma
from fnmatch import fnmatch
from os import listdir
from os.path import isfile, join, isdir
import re
from typing import Generator, TextIO, List

from openvariant.annotation.annotation import Annotation
from openvariant.annotation.parser import AnnotationTypesParsers
from openvariant.config.config_annotation import AnnotationFormat, AnnotationGeneralKeys, ExcludesKeys, \
    AnnotationDelimiter, AnnotationTypes
from openvariant.utils.logger import log


def _open_file(file_path: str, mode='rt') -> TextIO:
    open_method = open
    if file_path.endswith('gz') or file_path.endswith('bgz'):
        open_method = gzip.open
    elif file_path.endswith('xz'):
        open_method = lzma.open

    return open_method(file_path, mode)


def _base_parser(lines: TextIO, delimiter: str) -> Generator[int, str, None]:
    try:
        read_tsv = csv.reader(lines, delimiter=AnnotationDelimiter[delimiter].value)
    except KeyError:
        raise KeyError(f"'{delimiter}' key not found.")
    for l_num, line in enumerate(read_tsv):  # lines, start=1):
        # Skip empty lines
        if len(line) == 0:
            print(line)
            continue

        # Skip comments
        if (line[0].startswith('#') or line[0].startswith('##') or line[0].startswith('browser') or line[0].startswith(
                'track')) \
                and not line[0].startswith('#CHROM'):
            continue

        yield l_num, line


def _parse_row(ann: dict, line: List, header: List, original_header: List, path: str) -> dict:
    annotations = ann[AnnotationGeneralKeys.ANNOTATION.name]
    row_parser = []
    remain_annotation = {}
    annot_plugins = []
    for k, v in annotations.items():
        try:
            value = float('nan')
            if v[0] == AnnotationTypes.PLUGIN.name:
                annot_plugins.append(v)
            elif v[0] != AnnotationTypes.MAPPING.name:
                value = AnnotationTypesParsers[v[0]].value(v, line, original_header, path)
            else:
                remain_annotation[k] = v
            row_parser.append(value)
        except IndexError:
            row_parser.append(float('nan'))

    row_parser = list(map(str, row_parser))
    dict_line = {h: row_parser[i] for i, h in enumerate(header)}

    for v in annot_plugins:
        dict_line = AnnotationTypesParsers[v[0]].value(v, line, original_header, path, dict_line)

    for k, v in remain_annotation.items():
        dict_line[k] = AnnotationTypesParsers[v[0]].value(v, line, original_header, path, dict_line)

    try:
        for k, v in dict_line.items():
            dict_line[k] = v.format(**dict_line)
    except AttributeError as e:
        raise AttributeError(f'Parsing annotations error for {dict_line}: {e}')

    return dict_line


def _apply_exclude(line: dict, excludes: List) -> bool:
    for exclude in excludes:
        try:
            value_line = line[exclude[ExcludesKeys.FIELD.value]]
            value_exclude = str(exclude[ExcludesKeys.VALUE.value])
            if value_exclude.startswith("!"):
                if value_line != value_exclude[1::]:
                    return True
            else:
                if value_line == value_exclude:
                    return True
        except (KeyError, ValueError):
            return False
    return False

def _parser(file: str, annotation: dict, delimiter: str, columns: List, excludes: List, group_by=None, display_header=True) -> \
        Generator[dict, None, None]:
    row = None
    fd = _open_file(file, "rt")

    header = list(annotation[AnnotationGeneralKeys.ANNOTATION.name].keys())
    original_header = None
    for lnum, line in _base_parser(fd, delimiter):
        if original_header is None:
            original_header = line
            try:
                if not display_header:
                    continue
                row = header

            except (ValueError, KeyError) as e:
                log.error(f"Error parsing header {e}")
        else:
            try:
                row = _parse_row(annotation, line, header, original_header, file)

                row_aux = {}
                if _apply_exclude(row, excludes):
                    continue

                if len(columns) != 0:
                    if group_by is not None and group_by not in columns:
                        row_aux[group_by] = row[group_by]
                    for col in columns:
                        row_aux[col] = row[col]

                    row = row_aux

            except (ValueError, IndexError, KeyError) as e:
                raise ValueError(f"Error parsing line: {lnum} {file}: {e}")

        yield row
    fd.close()


def _check_extension(ext: str, path: str) -> bool:
    match: bool = True
    if ext[0] == '*':
        match = fnmatch(path, ext)
    else:
        reg_apply = re.compile(ext + '$')
        match = len(reg_apply.findall(path)) != 0
    return match


def _extract_header(annotation: Annotation):
    return list(annotation.annotations.keys())


class Variant:

    def __init__(self, path: str, ann: Annotation) -> None:
        csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))
        self._path: str = path
        self._annotation: Annotation = ann
        self._header: List[str] = _extract_header(ann)

    def _unify(self, base_path: str, annotation: Annotation, group_by=None, display_header=True) -> Generator[dict, None, None]:
        an = annotation.structure
        if isfile(base_path):
            for ext, ann in an.items():
                if _check_extension(ext, base_path):
                    for x in _parser(base_path, ann, annotation.delimiter, annotation.columns, annotation.excludes,
                                     group_by, display_header):
                        display_header = False
                        yield x
        else:
            try:
                for file in listdir(base_path):
                    file_path = join(base_path, file)
                    if isfile(file_path):
                        for ext, ann in an.items():
                            if _check_extension(ext, file_path):
                                for x in _parser(file_path, ann, annotation.delimiter, annotation.columns, annotation.excludes,
                                                 group_by, display_header):
                                    display_header = False
                                    yield x
                    else:
                        for x in self._unify(file_path, annotation, display_header=display_header):
                            display_header = False
                            yield x
            except PermissionError as e:
                print(e)

    def read(self, group_key=None) -> Generator[dict, None, None]:
        for i, line in enumerate(self._unify(self._path, self._annotation, group_by=group_key)):
            if i != 0:
                yield line

    def save(self, file_path: str, display_header=True):
        if isdir(file_path):
            raise ValueError("The path must be a file.")
        with open(file_path, "w") as file:
            writer = csv.writer(file, delimiter=AnnotationFormat[self._annotation.format.upper()].value)
            for i, line in enumerate(self._unify(self._path, self._annotation)):
                if display_header and i == 0:
                    writer.writerow(line)
                elif i != 0:
                    writer.writerow(line)
            file.close()

    @property
    def path(self) -> str:
        return self._path

    @property
    def header(self) -> List[str]:
        return self._header

    #@property
    #def generator(self) -> Generator[dict, None, None]:
    #    return self._generator

    @property
    def annotation(self) -> Annotation:
        return self._annotation



from enum import Enum

ANNOTATION_EXTENSION = "yaml"
DEFAULT_FORMAT = 'TSV'
DEFAULT_COLUMNS = []
DEFAULT_RECURSIVE = False
DEFAULT_DELIMITER = 'T'


class AnnotationGeneralKeys(Enum):
    PATTERN = 'pattern'
    RECURSIVE = 'recursive'
    FORMAT = 'format'
    DELIMITER = 'delimiter'
    COLUMNS = 'columns'
    ANNOTATION = 'annotation'
    EXCLUDE = 'exclude'


class AnnotationKeys(Enum):
    TYPE = 'type'
    FIELD = 'field'
    PLUGIN = 'plugin'
    VALUE = 'value'
    FIELD_SOURCE = 'fieldSource'
    FUNCTION = 'function'
    REGEX = 'regex'
    FILE_MAPPING = 'fileMapping'
    FIELD_MAPPING = 'fieldMapping'
    FIELD_VALUE = 'fieldValue'


class ExcludesKeys(Enum):
    FIELD = 'field'
    VALUE = 'value'


class AnnotationTypes(Enum):
    STATIC = 'static'
    INTERNAL = 'internal'
    DIRNAME = 'dirname'
    FILENAME = 'filename'
    PLUGIN = 'plugin'
    MAPPING = 'mapping'


class AnnotationDelimiter(Enum):
    T = "\t"
    C = ","


class AnnotationFormat(Enum):
    TSV = "\t"
    CSV = ","



from collections import defaultdict
from functools import partial
from multiprocessing import Pool
from os import cpu_count
from subprocess import PIPE, Popen
from typing import Set, Generator, List, Tuple

from tqdm import tqdm

from openvariant.annotation.annotation import Annotation
from openvariant.find.find import find_files
from openvariant.utils.logger import log
from openvariant.utils.where import skip, parse_where
from openvariant.variant.variant import Variant


def _get_unique_values(file_path: str, annotation: Annotation, key: str) -> Set:
    values = set()
    result = Variant(file_path, annotation)

    try:
        for r in result.read(key):
            values.add(r[key])
    except KeyError:
        log.warn(f"'{key}' key not found in '{file_path}' file")
    return values


def group(base_path: str, annotation_path: str or None, key_by: str) -> Generator[str, List, None]:
    results = defaultdict(list)
    for file, ann in find_files(base_path, annotation_path):
        by_value = ann.annotations.get(key_by, None)

        if isinstance(by_value, tuple):
            values = _get_unique_values(file, ann, key_by)

            for s in values:
                a_clone = ann
                results[s].append((file, a_clone))

    for key, group_select in results.items():
        yield key, group_select


def _group_by_task(selection, where=None, key_by=None, script='', header=False) -> Tuple[str, List, bool]:
    where_clauses = parse_where(where)
    group_key, group_values = selection

    output = []
    if script is None:
        try:
            for value in group_values:
                result = Variant(value[0], value[1])

                columns = result.annotation.columns if len(result.annotation.columns) != 0 else result.header
                if header:
                    line = "\t".join([str(h).strip() for h in columns])
                    output.append(f"{line}")
                    header = False
                for row in result.read(key_by):
                    if skip(row, where_clauses):
                        continue
                    if row[key_by] == group_key:
                        line = "\t".join([str(row[h]).strip() for h in columns])
                        output.append(f"{line}")
        except BrokenPipeError:
            pass
        return group_key, output, False

    else:
        process = Popen(script, shell=True, stdin=PIPE, stdout=PIPE,
                   env={"GROUP_KEY": 'None' if group_key is None else group_key})
        try:
            for value in group_values:
                result = Variant(value[0], value[1])
                columns = result.annotation.columns if len(result.annotation.columns) != 0 else result.header

                if header:
                    line = "\t".join([str(h).strip() for h in columns])
                    process.stdin.write(f"{line}\n".encode())
                    process.stdin.flush()
                    header = False
                for row in result.read(key_by):
                    if skip(row, where_clauses):
                        continue
                    try:
                        if row[key_by] == group_key:
                            line = "\t".join([str(row[h]).strip() for h in columns])
                            process.stdin.write(f"{line}\n".encode())
                            process.stdin.flush()
                    except KeyError:
                        pass
            process.stdin.close()
        except BrokenPipeError:
            pass

        try:
            while True:
                out = process.stdout.readline().decode().strip()
                if out == "":
                    break
                output.append(out)
            process.stdout.close()
        except BrokenPipeError:
            pass
        return group_key, output, True


def group_by(base_path: str, annotation_path: str or None, script: str or None, key_by: str, where=None,
             cores=cpu_count(), quite=False, header: bool = False) -> Generator[str, List, bool]:
    selection = list(group(base_path, annotation_path, key_by))

    with Pool(cores) as pool:
        task = partial(_group_by_task, where=where, key_by=key_by,
                       script=script, header=header)  # , where=where_parsed, columns=columns, print_headers=headers)
        map_method = map if cores == 1 or len(selection) <= 1 else pool.imap_unordered

        for group_key, group_result, command in tqdm(
                map_method(task, selection),
                total=len(selection),
                desc="Computing groups".rjust(40),
                disable=(len(selection) < 2 or quite)
        ):
            yield group_key, group_result, command



from typing import List

from openvariant.config.config_annotation import AnnotationFormat


def format_line(line: List[str], out_format: str) -> str:
    return AnnotationFormat[out_format.upper()].value.join(line)



from openvariant.find.find import find_files
from openvariant.utils.format_line import format_line
from openvariant.utils.where import parse_where, skip
from openvariant.variant.variant import Variant


def cat(base_path: str, annotation_path: str or None, where: str = None, header_show: bool = True) -> None:
    for file, annotation in find_files(base_path, annotation_path):
        where_clauses = parse_where(where)
        result = Variant(base_path, annotation)
        header = result.header if len(annotation.columns) == 0 else annotation.columns
        if header_show:
            print(format_line(header, result.annotation.format))
            header_show = False
        for i, r in enumerate(result.read()):
            if isinstance(r, dict):
                if skip(r, where_clauses):
                    continue
                print(format_line(list(map(str, r.values())), result.annotation.format))





import glob
from fnmatch import fnmatch
from os import listdir
from os.path import isfile, join, basename, abspath, dirname
from typing import Generator
from copy import deepcopy

from openvariant.annotation.annotation import Annotation

from openvariant.config.config_annotation import ANNOTATION_EXTENSION


def _get_annotation_file(annotation: Annotation or None, file_name: str, file_path: str, base_path: str) -> \
        Generator[str, Annotation, None]:
    if annotation is None:
        for annotation_file in glob.iglob(join(dirname(abspath(base_path)), "*.{}".format(ANNOTATION_EXTENSION))):
            annotation = Annotation(annotation_file)

    ann = None
    pattern_matches = 0
    if annotation is not None:
        for pattern in annotation.patterns:
            if fnmatch(file_name, pattern):
                pattern_matches += 1
                ann = deepcopy(annotation)

    # Process filename and dirname annotations
    if ann is not None and pattern_matches > 0:  # and __where_match(file_annotations, where=where):
        yield file_path, ann


def _find_files(base_path: str, annotation: Annotation or None, fix: bool) -> Generator[str, Annotation, None]:

    if isfile(base_path):
        for f, a in _get_annotation_file(annotation, basename(base_path), base_path, base_path):
            yield f, a

    # Find files
    else:
        try:
            for file_name in listdir(base_path):
                file_path = join(base_path, file_name)
                if isfile(file_path):
                    # local_annotation = global_annotation
                    if not fix:
                        for annotation_file in glob.iglob(join(base_path, "*.{}".format(ANNOTATION_EXTENSION))):
                            annotation = Annotation(annotation_file)
                            # recursive = DEFAULT_RECURSIVE if global_annotation.recursive is None \
                            #    else global_annotation.recursive
                            # if recursive:
                            #    local_annotation = merge_annotations_structure(local_annotation, ann)
                    for f, a in _find_files(file_path, annotation, fix):
                        yield f, a
                else:
                    for annotation_file in glob.iglob(join(base_path, "*.{}".format(ANNOTATION_EXTENSION))):
                        annotation = Annotation(annotation_file)
                    # Search subfolders
                    for f, a in _find_files(file_path, annotation, fix):
                        yield f, a
        except PermissionError as e:
            print(e)


def find_files(base_path: str, annotation_path: str or None = None) -> Generator[str, Annotation, None]:
    annotation, fix = (Annotation(annotation_path), True) if annotation_path is not None else (None, False)
    return _find_files(base_path, annotation, fix)



import logging
import re

from typing import List
from yaml import safe_load, YAMLError

from openvariant.annotation.builder import AnnotationTypesBuilders
from openvariant.config.config_annotation import (AnnotationGeneralKeys,
                                                  AnnotationKeys,
                                                  AnnotationTypes,
                                                  ExcludesKeys,
                                                  DEFAULT_FORMAT,
                                                  DEFAULT_DELIMITER,
                                                  DEFAULT_COLUMNS, AnnotationFormat, AnnotationDelimiter)


def _read_annotation_file(path: str) -> dict:
    with open(path, 'r') as stream:
        try:
            return safe_load(stream)
        except YAMLError as exc:
            logging.error(exc)
        stream.close()


def _check_general_keys(annot: dict) -> None:
    # Pattern key
    if AnnotationGeneralKeys.PATTERN.value not in annot or not isinstance(
            annot[AnnotationGeneralKeys.PATTERN.value], list) \
            and not all(isinstance(x, str) for x in annot[AnnotationGeneralKeys.PATTERN.value]):
        raise KeyError(f"'{AnnotationGeneralKeys.PATTERN.value}' key not found or is not a str.")

    # Recursive key
    if AnnotationGeneralKeys.RECURSIVE.value in annot and \
            not isinstance(annot[AnnotationGeneralKeys.RECURSIVE.value], bool):
        raise KeyError(f"'{AnnotationGeneralKeys.RECURSIVE.value}' key is not a boolean.")

    # Format key
    if AnnotationGeneralKeys.FORMAT.value in annot and \
            (not isinstance(annot[AnnotationGeneralKeys.FORMAT.value], str) or
             annot[AnnotationGeneralKeys.FORMAT.value].upper() not in [e.name for e in AnnotationFormat]):
        raise KeyError(f"'{AnnotationGeneralKeys.FORMAT.value}' key is not a string.")

    # Delimiter key
    if AnnotationGeneralKeys.DELIMITER.value in annot and \
            (not isinstance(annot[AnnotationGeneralKeys.DELIMITER.value], str) or
             annot[AnnotationGeneralKeys.DELIMITER.value].upper() not in [e.name for e in AnnotationDelimiter]):
        raise KeyError(f"'{AnnotationGeneralKeys.DELIMITER.value}' key is not valid or is not a string.")

    # Columns key
    if AnnotationGeneralKeys.COLUMNS.value in annot and \
            not isinstance(annot[AnnotationGeneralKeys.COLUMNS.value], list):
        raise KeyError(f"'{AnnotationGeneralKeys.COLUMNS.value}' key is not a list.")

    # Annotations key
    if AnnotationGeneralKeys.ANNOTATION.value in annot and \
            not isinstance(annot[AnnotationGeneralKeys.ANNOTATION.value], list):
        raise KeyError(f"'{AnnotationGeneralKeys.ANNOTATION.value}' key is not a list.")

    # Excludes key
    if AnnotationGeneralKeys.EXCLUDE.value in annot and \
            (not isinstance(annot[AnnotationGeneralKeys.EXCLUDE.value], list) or
             not all([ExcludesKeys.FIELD.value in x and ExcludesKeys.VALUE.value in x
                      for x in annot[AnnotationGeneralKeys.EXCLUDE.value]])):
        raise KeyError(f"'{AnnotationGeneralKeys.EXCLUDE.value}' key in bad format.")


def _check_annotation_keys(annot: dict) -> None:
    # Type key
    if AnnotationKeys.TYPE.value not in annot or not isinstance(annot[AnnotationKeys.TYPE.value], str):
        raise KeyError(f"'{AnnotationKeys.TYPE.value}' key not found or is not a str.")
    if annot[AnnotationKeys.TYPE.value] not in [e.value for e in AnnotationTypes]:
        raise ValueError(f"'{AnnotationKeys.TYPE.value}' value is wrong.")

    # Field key
    if AnnotationKeys.FIELD.value not in annot or not isinstance(annot[AnnotationKeys.FIELD.value], str):
        raise KeyError(f"'{AnnotationKeys.FIELD.value}' key not found or is not a str.")

    # Value key
    # if (annot[AnnotationKeys.TYPE.value] == AnnotationTypes.STATIC.value or
    #    annot[AnnotationKeys.TYPE.value] == AnnotationTypes.INTERNAL.value) and \
    #        not isinstance(annot[AnnotationKeys.VALUE.value], str):
    #    raise KeyError(f"'{AnnotationKeys.VALUE.value}' key not found or is not a str.")

    # Field source key
    if (annot[AnnotationKeys.TYPE.value] == AnnotationTypes.INTERNAL.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.PLUGIN.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.MAPPING.value) and \
            AnnotationKeys.FIELD_SOURCE.value in annot and \
            not isinstance(annot[AnnotationKeys.FIELD_SOURCE.value], list):
        raise KeyError(f"'{AnnotationKeys.FIELD_SOURCE.value}' key not found or is not a list.")

    # Dirname and filename key
    if (annot[AnnotationKeys.TYPE.value] == AnnotationTypes.DIRNAME.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.FILENAME.value or
        annot[AnnotationKeys.TYPE.value] == AnnotationTypes.INTERNAL.value) and \
            AnnotationKeys.FUNCTION.value in annot and \
            re.compile("lambda[' ']+[a-zA-Z0-9]+[' ']*:[' ']*.*").search(annot[AnnotationKeys.FUNCTION.value]) is None:
        raise ValueError(f"'{AnnotationKeys.FUNCTION.value}' value is not an appropriated lambda function.")

    # Plugin key
    if annot[AnnotationKeys.TYPE.value] == AnnotationTypes.PLUGIN.value and \
            AnnotationKeys.PLUGIN.value not in annot:
        raise KeyError(f"'{AnnotationKeys.PLUGIN.value}' key not found.")

    if annot[AnnotationKeys.TYPE.value] == AnnotationTypes.PLUGIN.value and \
            (AnnotationKeys.PLUGIN.value in annot and not isinstance(annot[AnnotationKeys.PLUGIN.value], str)):
        raise ValueError(f"'{AnnotationKeys.PLUGIN.value}' is not a str.")

    # Mapping keys
    if annot[AnnotationKeys.TYPE.value] == AnnotationTypes.MAPPING.value and \
            (AnnotationKeys.FIELD_SOURCE.value not in annot or
             AnnotationKeys.FIELD_MAPPING.value not in annot or
             AnnotationKeys.FILE_MAPPING.value not in annot or
             AnnotationKeys.FIELD_VALUE.value not in annot or
             not isinstance(annot[AnnotationKeys.FIELD_SOURCE.value], list) or
             not isinstance(annot[AnnotationKeys.FIELD_MAPPING.value], str) or
             not isinstance(annot[AnnotationKeys.FILE_MAPPING.value], str) or
             not isinstance(annot[AnnotationKeys.FIELD_VALUE.value], str)):
        raise KeyError(
            f"'{AnnotationTypes.MAPPING.value}' not annotated well.")


class Annotation:
    _builders: dict = {}

    def __init__(self, path: str) -> None:
        self._builders: dict = {}

        self._register_builders()
        self._path = path
        raw_annotation = _read_annotation_file(path)
        _check_general_keys(raw_annotation)
        for annot in raw_annotation.get(AnnotationGeneralKeys.ANNOTATION.value, []):
            _check_annotation_keys(annot)

        patterns = raw_annotation[AnnotationGeneralKeys.PATTERN.value]
        self._patterns = patterns if isinstance(patterns, List) else [patterns]
        self._recursive = raw_annotation.get(AnnotationGeneralKeys.RECURSIVE.value, True)
        self._delimiter = raw_annotation.get(AnnotationGeneralKeys.DELIMITER.value, DEFAULT_DELIMITER).upper()
        self._columns = raw_annotation.get(AnnotationGeneralKeys.COLUMNS.value, DEFAULT_COLUMNS)

        self._format = raw_annotation.get(AnnotationGeneralKeys.FORMAT.value, DEFAULT_FORMAT).replace('.', '')
        self._excludes = raw_annotation.get(AnnotationGeneralKeys.EXCLUDE.value, [])

        self._annotations: dict = {}
        for k in raw_annotation.get(AnnotationGeneralKeys.ANNOTATION.value, []):
            self._annotations[k[AnnotationKeys.FIELD.value]] = \
                AnnotationTypesBuilders[k[AnnotationKeys.TYPE.value].upper()].value(k, self._path)

        self._check_columns()

    def _register_builders(self) -> None:
        for b in AnnotationTypes:
            self._builders[b.value] = AnnotationTypesBuilders[b.name].value

    def _check_columns(self) -> None:
        for col in self._columns:
            if col not in self._annotations:
                raise KeyError(f"'{col}' column unable to find.")

    # @property
    # def recursive(self) -> bool:
    #    return self._recursive

    @property
    def patterns(self) -> List[str]:
        return self._patterns

    @property
    def format(self) -> str:
        return self._format

    @property
    def delimiter(self) -> str:
        return self._delimiter

    @property
    def columns(self) -> List:
        return self._columns

    @property
    def annotations(self) -> dict:
        return self._annotations

    @property
    def excludes(self) -> List:
        return self._excludes

    @property
    def structure(self) -> dict:

        structure_aux = {AnnotationGeneralKeys.ANNOTATION.name: self._annotations,
                         AnnotationGeneralKeys.EXCLUDE.name: self._excludes}
        return {e: structure_aux for e in self._patterns}


'''
def merge_annotations_structure(ann_a: Annotation, ann_b: Annotation) -> Annotation:
    """
    :param ann_a: The first Annotation. This annotation
    has preference and will override B annotation if there is a conflict
    :param ann_b: The second Annotation.
    :return: The merge of A and B annotation
    """
    if ann_a is None:
        return copy.deepcopy(ann_b)
    elif ann_b is None:
        return copy.deepcopy(ann_a)

    ann_aa = copy.deepcopy(ann_a)

    ann_aa.set_patterns(list(set(ann_aa.patterns).union(set(ann_b.patterns))))

    excludes_total = ann_aa.excludes

    for k in ann_b.excludes:
        if k not in excludes_total:
            excludes_total.append(k)

    ann_aa.set_excludes(excludes_total)

    aa = {k: v for k, v in ann_aa.annotations.items()}
    for k, v in ann_b.annotations.items():
        if k not in list(aa.keys()):
            aa[k] = v

    ann_aa.set_annotations(aa)
    return ann_aa
'''



import re
from enum import Enum
from functools import partial
from os.path import basename, dirname, abspath, isfile, isdir
from typing import List

from openvariant.annotation.builder import Builder, StaticBuilder, InternalBuilder, FilenameBuilder, DirnameBuilder, \
    PluginBuilder, MappingBuilder


def _get_text_from_header(x: List, line: List, original_header: List, func: Builder or None) -> float or str:
    value = None
    for y in x:
        value = line[original_header.index(y)] if y in original_header else None
        try:
            value = func(value) if func is not None else value
        except AttributeError:
            value = None
        if value is not None:
            break
    if value is None:
        return float('nan')
    return value


def _static_parser(x: StaticBuilder, line: List = None, original_header: List = None, path: str = None) -> str:
    return str(x[1]) if x[1] is not None else str(float('nan'))


def _internal_parser(x: InternalBuilder, line: List = None, original_header: List = None, path: str = None) -> str:
    try:
        for y in x[1]:
            if isinstance(y, list):
                dict_field = {}
                for field in y:
                    value = line[original_header.index(field)] if field in original_header else None
                    if value is not None:
                        dict_field[field] = value
                        continue
                    if value is None:
                        dict_field[field] = str(float('nan'))
                try:
                    return str(x[3].format(**dict_field))
                except KeyError:
                    continue
            if isinstance(y, str):
                return str(_get_text_from_header(x[1], line, original_header, x[2]))
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')
    except SyntaxError:
        raise SyntaxError(f'Unable to parser function lambda on {x[0]} annotation')

    return str(float('nan'))


def _filename_parser(x: FilenameBuilder, line: List = None, original_header: List = None, path: str = None) -> str:

    try:
        if isdir(path):
            raise FileNotFoundError('Unable to find a filename')

        func_result = x[1](basename(path))
        value = x[2].findall(func_result)[0]
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')
    except SyntaxError:
        raise SyntaxError(f'Unable to parser function lambda on {x[0]} annotation')
    except (re.error, IndexError):
        raise re.error(f'Wrong regex pattern on {x[0]} annotation')

    return str(value) if value is not None else str(float('nan'))


def _dirname_parser(x: DirnameBuilder, line: List = None, original_header: List = None, path: str = None) -> str:
    try:
        if isdir(path):
            raise FileNotFoundError('Unable to find a dirname')

        func_result = x[1](basename(dirname(abspath(path))))
        value = x[2].findall(func_result)[0]
    except TypeError:
        raise TypeError(f'Unable to parser {x[0]} annotation')
    except SyntaxError:
        raise SyntaxError(f'Unable to parser function lambda on {x[0]} annotation')
    except (re.error, IndexError):
        raise re.error(f'Wrong regex pattern on {x[0]} annotation')

    return str(value) if value is not None else str(float('nan'))


def _plugin_parser(x: PluginBuilder, line: List = None, original_header: List = None, path: str = None,
                   dict_line: dict = None) -> dict:
    try:
        if dict_line is None:
            dict_line = {}
        value = x[2](dict_line)
    except KeyError as e:
        raise KeyError(f'Something went wrong on the plugin: {e}')
    #if len(x[1]) != 0:
    #    value = _get_text_from_header(x[1], line, original_header, None)
    return value if value is not None else str(float('nan'))


def _mapping_parser(x: MappingBuilder, line: List = None, original_header: List = None, path: str = None, dict_line: dict = None) \
        -> str:
    if x[1] is None or x[2] is None or line is None or original_header is None or dict_line is None:
        raise ValueError('Unable to make mapping.')
    value = None
    for field in x[1]:
        value = line[original_header.index(field)] if field in original_header else None
        if value is None:
            try:
                k = dict_line[field]
                value = x[2][k]
            except KeyError:
                pass
                #raise KeyError(f"Enable to found '{field}' in the mapping file.")
        if value is not None:
            break

    return str(value) if value is not None else str(float('nan'))


class AnnotationTypesParsers(Enum):
    STATIC = partial(_static_parser)
    INTERNAL = partial(_internal_parser)
    FILENAME = partial(_filename_parser)
    DIRNAME = partial(_dirname_parser)
    PLUGIN = partial(_plugin_parser)
    MAPPING = partial(_mapping_parser)



import csv
import re
import glob
import gzip
import importlib
from enum import Enum
from functools import partial
from os.path import dirname
from typing import Tuple, Any, List, Callable

from openvariant.config.config_annotation import AnnotationKeys, AnnotationTypes


class Builder:
    func: str = None

    def __init__(self, func: str) -> None:
        self.func = func

    def __call__(self, x: Any) -> Any:
        return eval(self.func)(x)


StaticBuilder = Tuple[str, float or int or str]
InternalBuilder = Tuple[str, List, Builder, str or float]
DirnameBuilder = Tuple[str, Builder, re.Pattern]
FilenameBuilder = Tuple[str, Builder, re.Pattern]
PluginBuilder = Tuple[str, List, Callable]
MappingBuilder = Tuple[str, List, dict]


def _get_dirname_filename_attributes(x: dict) -> Tuple[Builder, re.Pattern]:
    func_apply = Builder("(lambda y: y)") if AnnotationKeys.FUNCTION.value not in x \
        else Builder(x[AnnotationKeys.FUNCTION.value])
    try:
        regex_apply = re.compile('(.*)') if AnnotationKeys.REGEX.value not in x or x[AnnotationKeys.REGEX.value] is None \
            else re.compile(x[AnnotationKeys.REGEX.value])
    except re.error as e:
        raise re.error(f'Wrong regex pattern: {e}')
    return func_apply, regex_apply


def _static_builder(x: dict, base_path: str = None) -> StaticBuilder:
    try:
        value = x[AnnotationKeys.VALUE.value]
    except KeyError:
        raise KeyError('Static annotation is wrong specified.')
    return AnnotationTypes.STATIC.name, value


def _internal_builder(x: dict, base_path: str = None) -> InternalBuilder:
    try:
        value = x[AnnotationKeys.VALUE.value]
    except KeyError:
        value = float('nan')

    return AnnotationTypes.INTERNAL.name, x[AnnotationKeys.FIELD_SOURCE.value], Builder("(lambda y: y)") \
        if AnnotationKeys.FUNCTION.value not in x or x[AnnotationKeys.FUNCTION.value] is None or \
           len(x[AnnotationKeys.FUNCTION.value]) == 2 else Builder(x[AnnotationKeys.FUNCTION.value]), value


def _dirname_builder(x: dict, base_path: str = None) -> DirnameBuilder:
    func_apply, regex_apply = _get_dirname_filename_attributes(x)

    return AnnotationTypes.DIRNAME.name, func_apply, regex_apply


def _filename_builder(x: dict, base_path: str = None) -> FilenameBuilder:
    func_apply, regex_apply = _get_dirname_filename_attributes(x)

    return AnnotationTypes.FILENAME.name, func_apply, regex_apply


def _plugin_builder(x: dict, base_path: str = None) -> PluginBuilder:
    try:
        mod = importlib.import_module(f".{x[AnnotationTypes.PLUGIN.value]}", package="plugins")
        func = getattr(mod, x[AnnotationTypes.PLUGIN.value])
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"Unable to found '{x[AnnotationTypes.PLUGIN.value]}' plugin.")
    field_sources = x[AnnotationKeys.FIELD_SOURCE.value] if AnnotationKeys.FIELD_SOURCE.value in x else []
    return AnnotationTypes.PLUGIN.name, field_sources, func


def _mapping_builder(x: dict, path: str) -> MappingBuilder:
    values: dict = {}
    mapping_files = x[AnnotationKeys.FILE_MAPPING.value]
    files = list(glob.iglob(dirname(path) + '/**/' + mapping_files, recursive=True))
    if len(files) == 0:
        raise FileNotFoundError(f'Unable to find \'{mapping_files}\' file in \'{path}\'')
    try:
        for mapping_file in files:
            open_method = gzip.open if mapping_file.endswith('gz') else open
            with open_method(mapping_file, "rt") as fd:
                for r in csv.DictReader(fd, delimiter='\t'):
                    field = r[x[AnnotationKeys.FIELD_MAPPING.value]]
                    val = r[x[AnnotationKeys.FIELD_VALUE.value]]
                    values[field] = val
            break
    except TypeError:
        raise TypeError("Unable to parse mapping annotation")
    return AnnotationTypes.MAPPING.name, x[AnnotationKeys.FIELD_SOURCE.value], values


class AnnotationTypesBuilders(Enum):
    STATIC = partial(_static_builder)
    INTERNAL = partial(_internal_builder)
    DIRNAME = partial(_dirname_builder)
    FILENAME = partial(_filename_builder)
    PLUGIN = partial(_plugin_builder)
    MAPPING = partial(_mapping_builder)
